/* 
 * tkWinMenu.c --
 *
 *	This module implements the Mac-platform specific features of menus.
 *
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tkWinMenu.c 1.60 96/12/17 15:18:28
 */

#define OEMRESOURCE
#include <string.h>
#include "tkMenu.h"
#include "tkWinInt.h"

/*
 * The class of the window for popup menus.
 */

#define MENU_CLASS_NAME "MenuWindowClass"

/*
 * Used to align a windows bitmap inside a rectangle
 */
#define ALIGN_BITMAP_LEFT   0x00000001
#define ALIGN_BITMAP_RIGHT  0x00000002
#define ALIGN_BITMAP_TOP    0x00000004
#define ALIGN_BITMAP_BOTTOM 0x00000008

static int indicatorDimensions[2];
				/* The dimensions of the indicator space
				 * in a menu entry. Calculated at init
				 * time to save time. */
static Tcl_HashTable commandTable;
				/* A map of command ids to menu entries */
static int inPostMenu;		/* We cannot be re-entrant like X Windows. */
static WORD lastCommandID;	/* The last command ID we allocated. */
static HWND menuHWND;		/* A window to service popup-menu messages
				 * in. */
static OSVERSIONINFO versionInfo;
				/* So we don't have to keep doing this */
static Tcl_HashTable winMenuTable;
				/* Need this to map HMENUs back to menuPtrs */

/*
 * The following are default menu value strings.
 */

static char borderString[5];	/* The string indicating how big the border is */
static Tcl_DString menuFontDString;
				/* A buffer to store the default menu font
				 * string. */

/*
 * Forward declarations for procedures defined later in this file:
 */

static void			DrawWindowsSystemBitmap _ANSI_ARGS_((
				    Display *display, Drawable drawable, 
				    GC gc, CONST RECT *rectPtr, int bitmapID,
				    int alignFlags));
static void			FreeID _ANSI_ARGS_((int commandID));
static char *			GetEntryText _ANSI_ARGS_((MenuEntry *mePtr));
static int			GetNewID _ANSI_ARGS_((MenuEntry *mePtr,
				    int *menuIDPtr));
static int			MenuKeyBindProc _ANSI_ARGS_((
				    ClientData clientData, 
				    Tcl_Interp *interp, XEvent *eventPtr,
				    Tk_Window tkwin, KeySym keySym));
static void			RedrawMenuBarWhenIdle _ANSI_ARGS_((
				    Menu *menuPtr));
static LRESULT CALLBACK		TkWinMenuProc _ANSI_ARGS_((HWND hwnd,
				    UINT message, WPARAM wParam,
				    LPARAM lParam));



/*
 *----------------------------------------------------------------------
 *
 * GetNewID --
 *
 *	Allocates a new menu id and marks it in use.
 *
 * Results:
 *	Returns TCL_OK if succesful; TCL_ERROR if there are no more
 *	ids of the appropriate type to allocate. menuIDPtr contains
 *	the new id if succesful.
 *
 * Side effects:
 *	An entry is created for the menu in the command hash table,
 *	and the hash entry is stored in the appropriate field in the
 *	menu data structure.
 *
 *----------------------------------------------------------------------
 */

static int
GetNewID(mePtr, menuIDPtr)
    MenuEntry *mePtr;		/* The menu we are working with */
    int *menuIDPtr;		/* The resulting id */
{
    int found = 0;
    int newEntry;
    Tcl_HashEntry *commandEntryPtr;
    WORD returnID;

    WORD curID = lastCommandID + 1;

    /*
     * The following code relies on WORD wrapping when the highest value is
     * incremented.
     */
    
    while (curID != lastCommandID) {
    	commandEntryPtr = Tcl_CreateHashEntry(&commandTable,
		(char *) curID, &newEntry);
    	if (newEntry == 1) {
    	    found = 1;
    	    returnID = curID;
    	    break;
    	}
    	curID++;
    }

    if (found) {
    	Tcl_SetHashValue(commandEntryPtr, (char *) mePtr);
    	*menuIDPtr = (int) returnID;
    	return TCL_OK;
    } else {
    	return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * FreeID --
 *
 *	Marks the itemID as free.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The hash table entry for the ID is cleared.
 *
 *----------------------------------------------------------------------
 */

static void
FreeID(commandID)
    int commandID;
{
    Tcl_HashEntry *entryPtr = Tcl_FindHashEntry(&commandTable,
	    (char *) commandID);
    
    if (entryPtr != NULL) {
    	 Tcl_DeleteHashEntry(entryPtr);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpNewMenu --
 *
 *	Gets a new blank menu. Only the platform specific options are filled
 *	in.
 *
 * Results:
 *	Returns a pointer to the menu structure. Not sure what to do when
 *	we can't get a Windows menu handle.
 *
 * Side effects:
 *	Allocates a Windows menu handle.
 *
 *----------------------------------------------------------------------
 */

PlatformData
TkpNewMenu(menuPtr)
    Menu *menuPtr;	/* The common structure we are making the
			 * platform structure for. */
{
    HMENU winMenuHdl;
    Tcl_HashEntry *hashEntryPtr;
    int newEntry;

    if (menuPtr->menuType == MENUBAR) {
	winMenuHdl = CreateMenu();
    } else {
	winMenuHdl = CreatePopupMenu();
    }

    /*
     * We hash all of the HMENU's so that we can get their menu ptrs
     * back when dispatch messages.
     */

    hashEntryPtr = Tcl_CreateHashEntry(&winMenuTable, (char *) winMenuHdl,
	    &newEntry);
    Tcl_SetHashValue(hashEntryPtr, (char *) menuPtr);

    return (PlatformData) winMenuHdl;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDestroyMenu --
 *
 *	Destroys platform-specific menu structures.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	All platform-specific allocations are freed up.
 *
 *----------------------------------------------------------------------
 */

void
TkpDestroyMenu(menuPtr)
    Menu *menuPtr;	    /* The common menu structure */
{
    HMENU winMenuHdl = (HMENU) menuPtr->platformData;

    DestroyMenu(winMenuHdl);
    menuPtr->platformData = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDestroyMenuEntry --
 *
 *	Cleans up platform-specific menu entry items.
 *
 * Results:
 *	None
 *
 * Side effects:
 *	All platform-specific allocations are freed up.
 *
 *----------------------------------------------------------------------
 */

void
TkpDestroyMenuEntry(mePtr)
    MenuEntry *mePtr;		    /* The entry to destroy */
{
    Menu *menuPtr = mePtr->menuPtr;
    HMENU winMenuHdl = (HMENU) menuPtr->platformData;
    
    RemoveMenu(winMenuHdl, (UINT) mePtr->platformEntryData, MF_BYCOMMAND);
    FreeID((int) mePtr->platformEntryData);
    mePtr->platformEntryData = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * GetEntryText --
 *
 *	Given a menu entry, gives back the text that should go in it.
 *	Separators should be done by the caller, as they have to be
 *	handled specially. Allocates the memory with alloc. The caller
 *	should free the memory.
 *
 * Results:
 *	itemText points to the new text for the item.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static char *
GetEntryText(mePtr)
    MenuEntry *mePtr;		/* A pointer to the menu entry. */
{
    char *itemText;

    if (mePtr->type == TEAROFF_ENTRY) {
	itemText = ckalloc(sizeof("(Tear-off)"));
	strcpy(itemText, "(Tear-off)");
    } else if (mePtr->imageString != NULL) {
	itemText = ckalloc(sizeof("(Image)"));
	strcpy(itemText, "(Image)");
    } else if (mePtr->bitmap != None) {
	itemText = ckalloc(sizeof("(Pixmap)"));
	strcpy(itemText, "(Pixmap)");
    } else if (mePtr->label == NULL || mePtr->labelLength == 0) {
	itemText = ckalloc(sizeof("( )"));
	strcpy(itemText, "( )");
    } else {
	int size = mePtr->labelLength + 1;
	int i, j;

	/*
	 * We have to construct the string with an ampersand
	 * preceeding the underline character, and a tab seperating
	 * the text and the accel text. We have to be careful with
	 * ampersands in the string.
	 */

	for (i = 0; i < mePtr->labelLength; i++) {
	    if (mePtr->label[i] == '&') {
		size++;
	    }
	}

	if (mePtr->underline >= 0) {
	    size++;
	}
	if (mePtr->label[mePtr->underline] == '&') {
	    size++;
	}

	if (mePtr->accelLength > 0) {
	    size += mePtr->accelLength + 1;
	}

	for (i = 0; i < mePtr->accelLength; i++) {
	    if (mePtr->accel[i] == '&') {
		size++;
	    }
	}

	itemText = ckalloc(size);
	
	if (mePtr->labelLength == 0) {
	    itemText[0] = 0;
	} else {
	    for (i = 0, j = 0; i < mePtr->labelLength; i++, j++) {
		if (mePtr->label[i] == '&') {
		    itemText[j++] = '&';
		}
		if (i == mePtr->underline) {
		    itemText[j++] = '&';
		}
		itemText[j] = mePtr->label[i];
	    }
	    itemText[j] = '\0';
	}

	if (mePtr->accelLength > 0) {
	    strcat(itemText, "\t");
	    for (i = 0, j = strlen(itemText); i < mePtr->accelLength;
		    i++, j++) {
		if (mePtr->accel[i] == '&') {
		    itemText[j++] = '&';
		}
		itemText[j] = mePtr->accel[i];
	    }
	    itemText[j] = '\0';
	}
    }
    return itemText;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpConfigureMenuEntry --
 *
 *	Processes configurations for menu entries.
 *
 * Results:
 *	Returns standard TCL result. If TCL_ERROR is returned, then
 *	interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information get set for mePtr; old resources
 *	get freed, if any need it.
 *
 *----------------------------------------------------------------------
 */

int
TkpConfigureMenuEntry(interp, menuPtr,  mePtr, index)
    Tcl_Interp *interp;			/* Used for error reporting. */
    Menu *menuPtr;			/* Information about whole menu. */
    register MenuEntry *mePtr;		/* Information about menu entry;  may
					 * or may not already have values for
					 * some fields. */
    int index;				/* Index of mePtr within menuPtr's
					 * entries. */
{
    HMENU winMenuHdl = (HMENU) menuPtr->platformData;

    char *itemText = NULL;
    LPCTSTR lpNewItem = NULL;
    UINT flags = MF_BYPOSITION;
    UINT itemID = 0;

    if (mePtr->type == SEPARATOR_ENTRY) {
	flags |= MF_SEPARATOR;
    } else {
	itemText = GetEntryText(mePtr);
	if (menuPtr->menuType == MENUBAR) {
	    lpNewItem = itemText;
	} else {
	    lpNewItem = (LPCTSTR) mePtr;
	    flags |= MF_OWNERDRAW;
	}

    	/*
    	 * Set enabling and disabling correctly.
    	 */

	if (mePtr->state == tkDisabledUid) {
	    flags |= MF_DISABLED;
	}
    	
    	/*
    	 * Set the check mark for check entries and radio entries.
    	 */
	
	if ((mePtr->type == CHECK_BUTTON_ENTRY
		|| mePtr->type == RADIO_BUTTON_ENTRY)
		&& mePtr->entryFlags & ENTRY_SELECTED) {
	    flags |= MF_CHECKED;
	}

	itemID = (int) mePtr->platformEntryData;
	if (mePtr->type == CASCADE_ENTRY) {
	    if ((mePtr->childMenuRefPtr != NULL)
		    && (mePtr->childMenuRefPtr->menuPtr != NULL)) {
		HMENU childMenuHdl = 
			(HMENU) mePtr->childMenuRefPtr->menuPtr->platformData;

		if (childMenuHdl == NULL) {
		    TkMenuCompleteIdleHandlers(mePtr->childMenuRefPtr
			    ->menuPtr);
		    childMenuHdl = (HMENU) mePtr->childMenuRefPtr->menuPtr
			    ->platformData;
		}
		if (childMenuHdl != NULL) {
		    itemID = (UINT) childMenuHdl;
		    flags |= MF_POPUP;
		}
	    }
	}
    }

    /*
     * ModifyMenu has a nasty bug in the Windows 3.x implementation.
     * Basically, if you change a cascade to a non-cascade, the next
     * menu call or two will crash. We have to remove the item and add
     * it again.
     */

    RemoveMenu(winMenuHdl, mePtr->index, MF_BYPOSITION);
    InsertMenu(winMenuHdl, mePtr->index, flags, itemID, lpNewItem);
    if (itemText != NULL) {
	ckfree(itemText);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpPostMenu --
 *
 *	Posts a menu on the screen
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The menu is posted and handled.
 *
 *----------------------------------------------------------------------
 */

int
TkpPostMenu(interp, menuPtr, x, y)
    Tcl_Interp *interp;
    Menu *menuPtr;
    int x;
    int y;
{
    HMENU winMenuHdl = (HMENU) menuPtr->platformData;
    int result;

    if (inPostMenu) {
        Tcl_AppendResult(interp,
		"Cannot call post menu while already posting menu",
		(char *) NULL);
    	result = TCL_ERROR;
    } else {
	RECT noGoawayRect;
	POINT point;
	Tk_Window parentWindow = Tk_Parent(menuPtr->tkwin);

	inPostMenu++;
    
	result = TkPreprocessMenu(menuPtr);
	if (result != TCL_OK) {
	    inPostMenu--;
	    return result;
	}

    	/*
    	 * The post commands could have deleted the menu, which means
    	 * we are dead and should go away.
    	 */
    	
    	if (menuPtr->tkwin == NULL) {
	    inPostMenu--;
    	    return TCL_OK;
    	}

	if (NULL == parentWindow) {
	    noGoawayRect.top = y - 50;
	    noGoawayRect.bottom = y + 50;
	    noGoawayRect.left = x - 50;
	    noGoawayRect.right = x + 50;
	} else {
	    int left, top;
	    Tk_GetRootCoords(parentWindow, &left, &top);
	    noGoawayRect.left = left;
	    noGoawayRect.top = top;
	    noGoawayRect.bottom = noGoawayRect.top + Tk_Height(parentWindow);
	    noGoawayRect.right = noGoawayRect.left + Tk_Width(parentWindow);
	}

	TrackPopupMenu(winMenuHdl, TPM_LEFTALIGN | TPM_LEFTBUTTON, x, y, 0, 
		menuHWND, &noGoawayRect);

	GetCursorPos(&point);
	Tk_PointerEvent(NULL, point.x, point.y);

	inPostMenu = 0;
	result = TCL_OK;
    }
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpMenuNewEntry --
 *
 *	Adds a pointer to a new menu entry structure with the platform-
 *	specific fields filled in.
 *
 * Results:
 *	Returns a pointer to the new entry.
 *
 * Side effects:
 *	Storage gets allocated.
 *
 *----------------------------------------------------------------------
 */

PlatformEntryData
TkpMenuNewEntry(mePtr)
    MenuEntry *mePtr;
{
    int commandID;
    Menu *menuPtr = mePtr->menuPtr;
    HMENU winMenuHdl = (HMENU) menuPtr->platformData;
    char *itemText = mePtr->label;
    char space[2] = " ";

    if ((itemText == NULL) || (itemText[0] == '\0')) {
	itemText = space;
    }

    GetNewID(mePtr, &commandID);
    mePtr->platformEntryData = (PlatformEntryData) commandID;

    InsertMenu(winMenuHdl, mePtr->index, MF_BYPOSITION, commandID, itemText);

    return (PlatformEntryData) commandID;
}

/*
 *----------------------------------------------------------------------
 *
 * TkWinMenuProc --
 *
 *	The window proc for the dummy window we put popups in. This allows
 *	is to post a popup whether or not we know what the parent window
 *	is.
 *
 * Results:
 *	Returns whatever is appropriate for the message in question.
 *
 * Side effects:
 *	Normal side-effect for windows messages.
 *
 *----------------------------------------------------------------------
 */

static LRESULT CALLBACK
TkWinMenuProc(hwnd, message, wParam, lParam)
    HWND hwnd;
    UINT message;
    WPARAM wParam;
    LPARAM lParam;
{
    LRESULT lResult;

    if (!TkWinHandleMenuEvent(&hwnd, &message, &wParam, &lParam, &lResult)) {
	lResult = DefWindowProc(hwnd, message, wParam, lParam);
    }
    return lResult;
}

/*
 *----------------------------------------------------------------------
 *
 * TkHandleMenuEvent --
 *
 *	Filters out menu messages from messages passed to a top-level.
 *	Will respond appropriately to WM_COMMAND, WM_MENUSELECT,
 *	WM_MEASUREITEM, WM_DRAWITEM
 *
 * Result:
 *	Returns 1 if this handled the message; 0 if it did not.
 *
 * Side effects:
 *	All of the parameters may be modified so that the caller can
 *	think it is getting a different message. plResult points to
 *	the result that should be returned to windows from this message.
 *
 *----------------------------------------------------------------------
 */

int
TkWinHandleMenuEvent(phwnd, pMessage, pwParam, plParam, plResult)
    HWND *phwnd;
    UINT *pMessage;
    WPARAM *pwParam;
    LPARAM *plParam;
    LRESULT *plResult;
{
    Tcl_HashEntry *hashEntryPtr;
    int returnResult = 0;
    Menu *menuPtr;
    MenuEntry *mePtr;

    switch (*pMessage) {
	case WM_INITMENU:
	    TkMenuInit();
	    hashEntryPtr = Tcl_FindHashEntry(&winMenuTable, (char *) *pwParam);
	    if (hashEntryPtr != NULL) {
		menuPtr = (Menu *) Tcl_GetHashValue(hashEntryPtr);
		TkPreprocessMenu(menuPtr);
		*plResult = 0;
		returnResult = 1;
	    }
	    break;

	case WM_COMMAND: {
	    TkMenuInit();
	    if (HIWORD(*pwParam) == 0) {
		hashEntryPtr = Tcl_FindHashEntry(&commandTable,
			(char *)LOWORD(*pwParam));
		if (hashEntryPtr != NULL) {
		    mePtr = (MenuEntry *) Tcl_GetHashValue(hashEntryPtr);
		    if (mePtr != NULL) {
		    	TkInvokeMenu(mePtr->menuPtr->interp,
				mePtr->menuPtr, mePtr->index);
		    }
		}
	    }
	    *plResult = 0;
	    returnResult = 1;
	    break;
	}


	case WM_MENUCHAR: {
	    unsigned char menuChar = (unsigned char) LOWORD(*pwParam);
	    hashEntryPtr = Tcl_FindHashEntry(&winMenuTable, (char *) *plParam);
	    if (hashEntryPtr != NULL) {
		int i;

		*plResult = 0;
		menuPtr = (Menu *) Tcl_GetHashValue(hashEntryPtr);
		for (i = 0; i < menuPtr->numEntries; i++) {
		    int underline = menuPtr->entries[i]->underline;
		    if ((-1 != underline) 
			    && (NULL != menuPtr->entries[i]->label)
			    && (CharUpper((LPTSTR) menuChar) 
			    == CharUpper((LPTSTR) (unsigned char) menuPtr
			    ->entries[i]->label[underline]))) {
			*plResult = (2 << 16) | i;
			break;
		    }
		}
		returnResult = 1;
	    }
	    break;
	}

	case WM_MEASUREITEM: {
	    LPMEASUREITEMSTRUCT itemPtr = (LPMEASUREITEMSTRUCT) *plParam;
    
	    if (itemPtr != NULL) {
		mePtr = (MenuEntry *) itemPtr->itemData;
		menuPtr = mePtr->menuPtr;

		TkRecomputeMenu(menuPtr);
		itemPtr->itemHeight = mePtr->height;
		itemPtr->itemWidth = menuPtr->indicatorSpace 
			+ menuPtr->labelWidth + menuPtr->accelWidth
			+ 2 * menuPtr->borderWidth 
			+ 2 * menuPtr->activeBorderWidth;
		*plResult = 1;
		returnResult = 1;
	    }
	    break;
	}
	
	case WM_DRAWITEM: {
	    TkWinDrawable *twdPtr;
	    LPDRAWITEMSTRUCT itemPtr = (LPDRAWITEMSTRUCT) *plParam;
	    Tk_FontMetrics fontMetrics;

	    if (itemPtr != NULL) {
		mePtr = (MenuEntry *) itemPtr->itemData;
		menuPtr = mePtr->menuPtr;
		twdPtr = (TkWinDrawable *) ckalloc(sizeof(TkWinDrawable));
		twdPtr->type = TWD_WINDC;
		twdPtr->winDC.hdc = itemPtr->hDC;

		if (itemPtr->itemState & ODS_SELECTED) {
		    TkActivateMenuEntry(menuPtr, mePtr->index);
		} else {
		    TkActivateMenuEntry(menuPtr, -1);
		}

		Tk_GetFontMetrics(menuPtr->tkfont, &fontMetrics);
		TkDrawMenuEntry(mePtr, (Drawable) twdPtr, menuPtr->tkfont,
			&fontMetrics, itemPtr->rcItem.left,
			itemPtr->rcItem.top, itemPtr->rcItem.right
			- itemPtr->rcItem.left, itemPtr->rcItem.bottom
			- itemPtr->rcItem.top, 0, 0);

		ckfree((char *) twdPtr);
		*plResult = 1;
		returnResult = 1;
	    }
	    break;
	}

    }
    return returnResult;
}
    

/*
 *----------------------------------------------------------------------
 *
 * TkpSetWindowMenuBar --
 *
 *	Associates a given menu with a window.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	On Windows and UNIX, associates the platform menu with the
 *	platform window.
 *
 *----------------------------------------------------------------------
 */

void
TkpSetWindowMenuBar(tkwin, menuPtr)
    Tk_Window tkwin;	    /* The window we are putting the menubar into.*/
    Menu *menuPtr;	    /* The menu we are inserting */
{
    HMENU winMenuHdl;
    
    if (menuPtr != NULL) {
	TkMenuCompleteIdleHandlers(menuPtr);
	winMenuHdl = (HMENU) menuPtr->platformData;
	TkWinSetMenu(tkwin, winMenuHdl);
    } else {
	TkWinSetMenu(tkwin, NULL);
    }
}


/*
 *----------------------------------------------------------------------
 *
 * TkpSetMainMenubar --
 *
 *	Puts the menu associated with a window into the menubar. Should
 *	only be called when the window is in front.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The menubar is changed.
 *
 *----------------------------------------------------------------------
 */
void
TkpSetMainMenubar(
    Tcl_Interp *interp,		/* The interpreter of the application */
    Tk_Window tkwin,		/* The frame we are setting up */
    char *menuName)		/* The name of the menu to put in front.
    				 * If NULL, use the default menu bar.
    				 */
{
    /*
     * Nothing to do.
     */
}

/*
 *----------------------------------------------------------------------
 *
 * TkpGetMenuIndicatorGeometry --
 *
 *	Gets the width and height of the indicator area of a menu.
 *
 * Results:
 *	widthPtr and heightPtr are set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkpGetMenuIndicatorGeometry (
    Menu *menuPtr,			/* The menu we are measuring */
    MenuEntry *mePtr,			/* The entry we are measuring */
    int *widthPtr,			/* The resulting width */
    int *heightPtr,			/* The resulting height */
    Tk_Font tkfont,			/* Precalculated font */
    CONST Tk_FontMetrics *fmPtr)	/* Precalculated font metrics */
{
    *heightPtr = indicatorDimensions[0];
    *widthPtr = indicatorDimensions[1];
}

/*
 *----------------------------------------------------------------------
 *
 * TkpGetMenuAccelGeometry --
 *
 *	Gets the width and height of the indicator area of a menu.
 *
 * Results:
 *	widthPtr and heightPtr are set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkpGetMenuAccelGeometry (
    Menu *menuPtr,			/* The menu we are measuring */
    MenuEntry *mePtr,			/* The entry we are measuring */
    int *widthPtr,			/* The resulting width */
    int *heightPtr,			/* The resulting height */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr)	/* The precalculated font metrics */
{
    *heightPtr = fmPtr->linespace;
    if (mePtr->type == CASCADE_ENTRY) {
	*widthPtr = 0;
    } else if (mePtr->accel == NULL) {
	*widthPtr = 0;
    } else {
	*widthPtr = Tk_TextWidth(tkfont, mePtr->accel, mePtr->accelLength);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpGetTearoffEntryGeometry --
 *
 *	Gets the width and height of the indicator area of a menu.
 *
 * Results:
 *	widthPtr and heightPtr are set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkpGetTearoffEntryGeometry (
    Menu *menuPtr,			/* The menu we are measuring */
    MenuEntry *mePtr,			/* The entry we are measuring */
    int *widthPtr,			/* The resulting width */
    int *heightPtr,			/* The resulting height */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr)	/* The precalculated font metrics */
{
    *heightPtr = fmPtr->linespace;
    *widthPtr = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpGetMenuSeparatorGeometry --
 *
 *	Gets the width and height of the indicator area of a menu.
 *
 * Results:
 *	widthPtr and heightPtr are set.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

void
TkpGetMenuSeparatorGeometry (
    Menu *menuPtr,			/* The menu we are measuring */
    MenuEntry *mePtr,			/* The entry we are measuring */
    int *widthPtr,			/* The resulting width */
    int *heightPtr,			/* The resulting height */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr)	/* The precalcualted font metrics */
{
    *widthPtr = 0;
    *heightPtr = fmPtr->linespace;
}

/*
 *----------------------------------------------------------------------
 *
 * DrawWindowsSystemBitmap --
 *
 *	Draws the windows system bitmap given by bitmapID into the rect
 *	given by rectPtr in the drawable. The bitmap is centered in the
 *	rectangle. It is not clipped, so if the bitmap is bigger than
 *	the rect it will bleed.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Drawing occurs. Some storage is allocated and released.
 *
 *----------------------------------------------------------------------
 */

static void
DrawWindowsSystemBitmap(display, drawable, gc, rectPtr, bitmapID, alignFlags)
    Display *display;			/* The display we are drawing into */
    Drawable drawable;			/* The drawable we are working with */
    GC gc;				/* The GC to draw with */
    CONST RECT *rectPtr;		/* The rectangle to draw into */			
    int bitmapID;			/* The windows id of the system
					 * bitmap to draw. */
    int alignFlags;			/* How to align the bitmap inside the
					 * rectangle. */
{
    TkWinDCState state;
    HDC hdc = TkWinGetDrawableDC(display, drawable, &state);
    HDC scratchDC;
    HBITMAP bitmap;
    BITMAP bm;
    POINT ptSize;
    POINT ptOrg;
    int topOffset, leftOffset;
    
    SetBkColor(hdc, gc->background);
    SetTextColor(hdc, gc->foreground);

    scratchDC = CreateCompatibleDC(hdc);
    bitmap = LoadBitmap(NULL, MAKEINTRESOURCE(bitmapID));

    SelectObject(scratchDC, bitmap);
    SetMapMode(scratchDC, GetMapMode(hdc));
    GetObject(bitmap, sizeof(BITMAP), &bm);
    ptSize.x = bm.bmWidth;
    ptSize.y = bm.bmHeight;
    DPtoLP(hdc, &ptSize, 1);

    ptOrg.y = ptOrg.x = 0;
    DPtoLP(hdc, &ptOrg, 1);

    if (alignFlags & ALIGN_BITMAP_TOP) {
	topOffset = 0;
    } else if (alignFlags & ALIGN_BITMAP_BOTTOM) {
	topOffset = (rectPtr->bottom - rectPtr->top) - ptSize.y;
    } else {
	topOffset = (rectPtr->bottom - rectPtr->top) / 2 - (ptSize.y / 2);
    }

    if (alignFlags & ALIGN_BITMAP_LEFT) {
	leftOffset = 0;
    } else if (alignFlags & ALIGN_BITMAP_RIGHT) {
	leftOffset = (rectPtr->right - rectPtr->left) - ptSize.x;
    } else {
	leftOffset = (rectPtr->right - rectPtr->left) / 2 - (ptSize.x / 2);
    }
    
    BitBlt(hdc, rectPtr->left + leftOffset, rectPtr->top + topOffset, ptSize.x,
	    ptSize.y, scratchDC, ptOrg.x, ptOrg.y, SRCCOPY);
    DeleteDC(scratchDC);
    DeleteObject(bitmap);

    TkWinReleaseDrawableDC(drawable, hdc, &state);
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuEntryIndicator --
 *
 *	This procedure draws the indicator part of a menu.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */
void
TkpDrawMenuEntryIndicator(menuPtr, mePtr, d, gc, indicatorGC, tkfont, fmPtr, x,
	y, width, height)
    Menu *menuPtr;		    /* The menu we are drawing */
    MenuEntry *mePtr;		    /* The entry we are drawing */
    Drawable d;			    /* What we are drawing into */
    GC gc;			    /* The gc we are drawing with */
    GC indicatorGC;		    /* The gc for indicator objects */
    Tk_Font tkfont;		    /* The precalculated font */
    CONST Tk_FontMetrics *fmPtr;    /* The precalculated font metrics */
    int x;			    /* Left edge */
    int y;			    /* Top edge */
    int width;
    int height;
{
    if ((mePtr->type == CHECK_BUTTON_ENTRY || 
    	    mePtr->type == RADIO_BUTTON_ENTRY) 
    	    && mePtr->entryFlags & ENTRY_SELECTED) {
	RECT rect;
	GC whichGC;

	if (mePtr->state == tkActiveUid) {
	    whichGC = gc;
	} else {
	    whichGC = indicatorGC;
	}

	rect.top = y;
	rect.bottom = y + mePtr->height;
	rect.left = menuPtr->borderWidth + menuPtr->activeBorderWidth + x;
	rect.right = menuPtr->indicatorSpace + x;
	DrawWindowsSystemBitmap(menuPtr->display, d, whichGC, &rect, 
		OBM_CHECK, 0);
    }    
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuEntryAccelerator --
 *
 *	This procedure draws the accelerator part of a menu. We
 *	need to decide what to draw here. Should we replace strings
 *	like "Control", "Command", etc?
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

void
TkpDrawMenuEntryAccelerator(menuPtr, mePtr, d, gc, tkfont, fmPtr,
	activeBorder, x, y, width, height, drawArrow)
    Menu *menuPtr;			/* The menu we are drawing */
    MenuEntry *mePtr;			/* The entry we are drawing */
    Drawable d;				/* What we are drawing into */
    GC gc;				/* The gc we are drawing with */
    Tk_Font tkfont;			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr;	/* The precalculated font metrics */
    Tk_3DBorder activeBorder;		/* The border when an item is active */
    int x;				/* left edge */
    int y;				/* top edge */
    int width;				/* Width of menu entry */
    int height;				/* Height of menu entry */
    int drawArrow;			/* For cascade menus, whether of not
					 * to draw the arraw. I cannot figure
					 * out Windows' algorithm for where
					 * to draw this. */
{
    int baseline;
    int leftEdge = x + menuPtr->indicatorSpace + menuPtr->labelWidth;

    baseline = y + (height + fmPtr->ascent - fmPtr->descent) / 2;
    if (mePtr->accel != NULL) {
	Tk_DrawChars(menuPtr->display, d, gc, tkfont, mePtr->accel, 
		mePtr->accelLength, leftEdge, baseline);
    }

    if ((mePtr->type == CASCADE_ENTRY) && drawArrow) {
	RECT rect;

	rect.top = y + GetSystemMetrics(SM_CYBORDER);
	rect.bottom = y + height - GetSystemMetrics(SM_CYBORDER);
	rect.left = x + menuPtr->indicatorSpace + menuPtr->labelWidth;
	rect.right = x + width - 1;
	DrawWindowsSystemBitmap(menuPtr->display, d, gc, &rect, OBM_MNARROW, 
		ALIGN_BITMAP_RIGHT);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawTearoffEntry --
 *
 *	This draws a tearoff entry. On the Mac, these are zero-sized
 *	and are not drawn.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */

void
TkpDrawTearoffEntry(menuPtr, mePtr, d, gc, tkfont, fmPtr, x, y, width, height)
    Menu *menuPtr;			/* The menu we are drawing */
    MenuEntry *mePtr;			/* The entry we are going to draw */
    Drawable d;				/* What we are drawing into */
    GC gc;				/* The gc we are going to use */
    Tk_Font tkfont;			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr;	/* The precalculated font metrics */
    int x;				/* left edge */
    int y;				/* top edge */
    int width;				/* Width of entry */
    int height;				/* Height of entry */
{
    XPoint points[2];
    int segmentWidth, maxX;
    Tk_FontMetrics fm;
    
    if (tkfont == NULL) {
        tkfont = menuPtr->tkfont;
    }
    Tk_GetFontMetrics(tkfont, &fm);
    
    points[0].x = x;
    points[0].y = y + height/2;
    points[1].y = points[0].y;
    segmentWidth = 6;
    maxX  = x + width - 1;

    while (points[0].x < maxX) {
	points[1].x = points[0].x + segmentWidth;
	if (points[1].x > maxX) {
	    points[1].x = maxX;
	}
	Tk_Draw3DPolygon(menuPtr->tkwin, d,
		menuPtr->border, points, 2, 1, TK_RELIEF_FLAT);
	points[0].x += 2 * segmentWidth;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuSeparator --
 *
 *	The menu separator is drawn.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */
void
TkpDrawMenuSeparator(menuPtr, mePtr, d, gc, tkfont, fmPtr, x, y, width, height)
    Menu *menuPtr;			/* The menu we are drawing */
    MenuEntry *mePtr;			/* The entry we are drawing */
    Drawable d;				/* What we are drawing into */
    GC gc;				/* The gc we are drawing with */
    Tk_Font tkfont;			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr;	/* The precalculated font metrics */
    int x;				/* left edge */
    int y;				/* top edge */
    int width;				/* width of item */
    int height;				/* height of item */
{
    XPoint points[2];

    points[0].x = x;
    points[0].y = y + height / 2;
    points[1].x = x + width - 1;
    points[1].y = points[0].y;
    Tk_Draw3DPolygon(menuPtr->tkwin, d,
	    menuPtr->border, points, 2, 1, TK_RELIEF_RAISED);
}

/*
 *----------------------------------------------------------------------
 *
 * TkpDrawMenuUnderline --
 *
 *	On appropriate platforms, draw the underline character for the
 *	menu.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Commands are output to X to display the menu in its
 *	current mode.
 *
 *----------------------------------------------------------------------
 */
void
TkpDrawMenuUnderline(
    Menu *menuPtr,			/* The menu to draw into */
    MenuEntry *mePtr,			/* The entry we are drawing */
    Drawable d,				/* What we are drawing into */
    GC gc,				/* The gc to draw into */
    Tk_Font tkfont,			/* The precalculated font */
    CONST Tk_FontMetrics *fmPtr,	/* The precalculated font metrics */
    int x,				/* Left Edge */
    int y,				/* Top Edge */
    int width,				/* Width of entry */
    int height)				/* Height of entry */
{
    if (mePtr->underline >= 0) {
    	Tk_UnderlineChars(menuPtr->display, d,
    		gc, tkfont, mePtr->label, x + menuPtr->indicatorSpace,
    		y + (height + fmPtr->ascent - fmPtr->descent) / 2, 
		mePtr->underline, mePtr->underline + 1);
    }		
}

/*
 *--------------------------------------------------------------
 *
 * MenuKeyBindProc --
 *
 *	This procedure is invoked when keys related to pulling
 *	down menus is pressed. The corresponding Windows events
 *	are generated and passed to DefWindowProc if appropriate.
 *
 * Results:
 *	Always returns TCL_OK.
 *
 * Side effects:
 *	The menu system may take over and process user events
 *	for menu input.
 *
 *--------------------------------------------------------------
 */

static int
MenuKeyBindProc(clientData, interp, eventPtr, tkwin, keySym)
    ClientData clientData;	/* not used in this proc */
    Tcl_Interp *interp;		/* The interpreter of the receiving window. */
    XEvent *eventPtr;		/* The XEvent to process */
    Tk_Window tkwin;		/* The window receiving the event */
    KeySym keySym;		/* The key sym that is produced. */
{
    UINT scanCode;
    UINT virtualKey;
    TkWindow *winPtr = (TkWindow *)tkwin;

    if (eventPtr->type == KeyPress) {
	switch (keySym) {
	case XK_Alt_L:
	    scanCode = MapVirtualKey(VK_LMENU, 0);
	    CallWindowProc(DefWindowProc, Tk_GetHWND(Tk_WindowId(tkwin)),
		    WM_SYSKEYDOWN, VK_MENU, (scanCode << 16)
		    | (1 << 29));
	    break;
	case XK_Alt_R:
	    scanCode = MapVirtualKey(VK_RMENU, 0);
	    CallWindowProc(DefWindowProc, Tk_GetHWND(Tk_WindowId(tkwin)),
		    WM_SYSKEYDOWN, VK_MENU, (scanCode << 16)
		    | (1 << 29) | (1 << 24));
	    break;
	case XK_F10:
	    scanCode = MapVirtualKey(VK_F10, 0);
	    CallWindowProc(DefWindowProc, Tk_GetHWND(Tk_WindowId(tkwin)),
		    WM_SYSKEYDOWN, VK_F10, (scanCode << 16));
	    break;
	default:
	    virtualKey = XKeysymToKeycode(winPtr->display, keySym);
	    scanCode = MapVirtualKey(virtualKey, 0);
	    if (0 != scanCode) {
		CallWindowProc(DefWindowProc, Tk_GetHWND(Tk_WindowId(tkwin)),
			WM_SYSKEYDOWN, virtualKey, ((scanCode << 16)
			| (1 << 29)));
	    }
	}
    } else if (eventPtr->type == KeyRelease) {
	switch (keySym) {
	case XK_Alt_L:
	    scanCode = MapVirtualKey(VK_LMENU, 0);
	    CallWindowProc(DefWindowProc, Tk_GetHWND(Tk_WindowId(tkwin)),
		    WM_SYSKEYUP, VK_MENU, (scanCode << 16)
		    | (1 << 29) | (1 << 30) | (1 << 31));
	    break;
	case XK_Alt_R:
	    scanCode = MapVirtualKey(VK_RMENU, 0);
	    CallWindowProc(DefWindowProc, Tk_GetHWND(Tk_WindowId(tkwin)),
		    WM_SYSKEYUP, VK_MENU, (scanCode << 16) | (1 << 24)
		    | (0x111 << 29) | (1 << 30) | (1 << 31));
	    break;
	case XK_F10:
	    scanCode = MapVirtualKey(VK_F10, 0);
	    CallWindowProc(DefWindowProc, Tk_GetHWND(Tk_WindowId(tkwin)),
		    WM_SYSKEYUP, VK_F10, (scanCode << 16)
		    | (1 << 30) | (1 << 31));
	    break;
	default:
	    virtualKey = XKeysymToKeycode(winPtr->display, keySym);
	    scanCode = MapVirtualKey(virtualKey, 0);
	    if (0 != scanCode) {
		CallWindowProc(DefWindowProc, Tk_GetHWND(Tk_WindowId(tkwin)),
			WM_SYSKEYUP, virtualKey, ((scanCode << 16)
			| (1 << 29) | (1 << 30) | (1 << 31)));
	    }
	}
    }
    return TCL_OK;
}   

/*
 *--------------------------------------------------------------
 *
 * TkpInitializeMenuBindings --
 *
 *	For every interp, initializes the bindings for Windows
 *	menus. Does nothing on Mac or XWindows.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	C-level bindings are setup for the interp which will
 *	handle Alt-key sequences for menus without beeping
 *	or interfering with user-defined Alt-key bindings.
 *
 *--------------------------------------------------------------
 */

void
TkpInitializeMenuBindings(interp, bindingTable)
    Tcl_Interp *interp;		    /* The interpreter to set. */
    Tk_BindingTable bindingTable;   /* The table to add to. */
{
    Tk_Uid uid = Tk_GetUid("all");

    /*
     * We need to set up the bindings for menubars. These have to
     * recreate windows events, so we need to have a C-level
     * binding for this. We have to generate the WM_SYSKEYDOWNS
     * and WM_SYSKEYUPs appropriately.
     */
    
    TkCreateBindingProcedure(interp, bindingTable, (ClientData)uid, 
	    "<Alt_L>", MenuKeyBindProc, NULL, NULL);
    TkCreateBindingProcedure(interp, bindingTable, (ClientData)uid,
	    "<KeyRelease-Alt_L>", MenuKeyBindProc, NULL, NULL);
    TkCreateBindingProcedure(interp, bindingTable, (ClientData)uid, 
	    "<Alt_R>", MenuKeyBindProc, NULL, NULL);
    TkCreateBindingProcedure(interp, bindingTable, (ClientData)uid,
	    "<KeyRelease-Alt_R>", MenuKeyBindProc, NULL, NULL);
    TkCreateBindingProcedure(interp, bindingTable, (ClientData)uid,
	    "<Alt-KeyPress>", MenuKeyBindProc, NULL, NULL);
    TkCreateBindingProcedure(interp, bindingTable, (ClientData)uid,
	    "<Alt-KeyRelease>", MenuKeyBindProc, NULL, NULL);
    TkCreateBindingProcedure(interp, bindingTable, (ClientData)uid,
	    "<KeyPress-F10>", MenuKeyBindProc, NULL, NULL);
    TkCreateBindingProcedure(interp, bindingTable, (ClientData)uid,
	    "<KeyRelease-F10>", MenuKeyBindProc, NULL, NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * TkpMenuInit --
 *
 *	Sets up the hash tables and the variables used by the menu package.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	lastMenuID gets initialized, and the parent hash and the command hash
 *	are allocated.
 *
 *----------------------------------------------------------------------
 */

void
TkpMenuInit()
{
    WNDCLASS wndClass;
    char sizeString[4];
    int pixelHeight;
    char faceName[LF_FACESIZE];
    HDC scratchDC;
    Tcl_DString boldItalicDString;
    int bold = 0; 
    int italic = 0;
    int i;
    TEXTMETRIC tm;

    Tcl_InitHashTable(&winMenuTable, TCL_ONE_WORD_KEYS);
    Tcl_InitHashTable(&commandTable, TCL_ONE_WORD_KEYS);
 
    wndClass.style = CS_OWNDC;
    wndClass.lpfnWndProc = TkWinMenuProc;
    wndClass.cbClsExtra = 0;
    wndClass.cbWndExtra = 0;
    wndClass.hInstance = Tk_GetHINSTANCE();
    wndClass.hIcon = NULL;
    wndClass.hCursor = NULL;
    wndClass.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wndClass.lpszMenuName = NULL;
    wndClass.lpszClassName = MENU_CLASS_NAME;
    RegisterClass(&wndClass);

    menuHWND = CreateWindow(MENU_CLASS_NAME, "MenuWindow", WS_POPUP,
	0, 0, 10, 10, NULL, NULL, Tk_GetHINSTANCE(), NULL);

    versionInfo.dwOSVersionInfoSize = sizeof(versionInfo);

    /*
     * If GetVersionEx fails, it means that the version info record
     * is too big for what is compiled. Should never happen, but if
     * it does, we are later than Windows 95 or NT 4.0.
     */

    if (!GetVersionEx(&versionInfo)) {
	versionInfo.dwMajorVersion = 4;
    }

    /*
     * Set all of the default options. The loop will terminate when we run 
     * out of options via a break statement.
     */

    for (i = 0; ; i++) {
	if (tkMenuConfigSpecs[i].type == TK_CONFIG_END) {
	    break;
	}

	if ((strcmp(tkMenuConfigSpecs[i].dbName,
		"activeBorderWidth") == 0) ||
		(strcmp(tkMenuConfigSpecs[i].dbName, "borderWidth") == 0)) {
	    int borderWidth;

	    borderWidth = GetSystemMetrics(SM_CXBORDER);
	    if (GetSystemMetrics(SM_CYBORDER) > borderWidth) {
		borderWidth = GetSystemMetrics(SM_CYBORDER);
	    }
	    sprintf(borderString, "%d", borderWidth);
	    tkMenuConfigSpecs[i].defValue = borderString;
	} else if ((strcmp(tkMenuConfigSpecs[i].dbName, "font") == 0)) {
	    HFONT menuFont;
	    scratchDC = CreateDC("DISPLAY", NULL, NULL, NULL);
	    Tcl_DStringInit(&menuFontDString);

	    if (versionInfo.dwMajorVersion >= 4) {
		NONCLIENTMETRICS ncMetrics;

		ncMetrics.cbSize = sizeof(ncMetrics);
		SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(ncMetrics),
			&ncMetrics, 0);
		pixelHeight = ncMetrics.lfMenuFont.lfHeight;
		menuFont = CreateFontIndirect(&ncMetrics.lfMenuFont);
	    } else {
		menuFont = GetStockObject(SYSTEM_FONT);
	    }
	    SelectObject(scratchDC, menuFont);
	    GetTextMetrics(scratchDC, &tm);
	    GetTextFace(scratchDC, sizeof(menuFontDString), faceName);
	    pixelHeight = tm.tmHeight;
	    if (tm.tmWeight >= 700) {
		bold = 1;
	    }
	    if (tm.tmItalic) {
		italic = 1;
	    }

	    DeleteDC(scratchDC);

	    if (versionInfo.dwMajorVersion >= 4) {
		DeleteObject(menuFont);
	    }
		
	    Tcl_DStringAppendElement(&menuFontDString, faceName);
	    sprintf(sizeString, "%d", pixelHeight);
	    Tcl_DStringAppendElement(&menuFontDString, sizeString);

	    if (bold == 1 || italic == 1) {
		Tcl_DStringInit(&boldItalicDString);
		if (bold == 1) {
		    Tcl_DStringAppendElement(&boldItalicDString, "bold");
		}
		if (italic == 1) {
		    Tcl_DStringAppendElement(&boldItalicDString, "italic");
		}
		Tcl_DStringAppendElement(&menuFontDString, 
			Tcl_DStringValue(&boldItalicDString));
	    }

	    tkMenuConfigSpecs[i].defValue = Tcl_DStringValue(&menuFontDString);
	}
    }

    /*
     * Now we go ahead and get the dimensions of the check mark and the
     * appropriate margins. Since this is fairly hairy, we do it here
     * to save time when traversing large sets of menu items.
     *
     * The code below was given to me by Microsoft over the phone. It
     * is the only way to insure menu items lining up, and is not
     * documented.
     */

    if (versionInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) {
	indicatorDimensions[0] = GetSystemMetrics(SM_CYMENUCHECK);
	indicatorDimensions[1] = ((GetSystemMetrics(SM_CXFIXEDFRAME) +
		GetSystemMetrics(SM_CXBORDER) 
		+ GetSystemMetrics(SM_CXMENUCHECK) + 7) & 0xFFF8)
		- GetSystemMetrics(SM_CXFIXEDFRAME);
    } else {
	DWORD dimensions = GetMenuCheckMarkDimensions();
	indicatorDimensions[0] = HIWORD(dimensions);
	indicatorDimensions[1] = LOWORD(dimensions);
   }

}

/*
 *--------------------------------------------------------------
 *
 * TkpComputeMenubarGeometry --
 *
 *	This procedure is invoked to recompute the size and
 *	layout of a menu that is a menubar clone.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Fields of menu entries are changed to reflect their
 *	current positions, and the size of the menu window
 *	itself may be changed.
 *
 *--------------------------------------------------------------
 */

void
TkpComputeMenubarGeometry(menuPtr)
    Menu *menuPtr;		/* Structure describing menu. */
{
    TkComputeStandardMenuGeometry(menuPtr);
    if (menuPtr->menuType == MENUBAR) {
	TkpSetWindowMenuBar(menuPtr->parentTopLevelPtr, menuPtr);
    }
}
